/* 
s.to.rast - this program was created to provided a command-line operability
of what could have been done interactively with the s.menu program. Thus,
much of what you find here is simply a modified version of the sites_to_rast
portion of the s.menu code.

This program reads a site file and creates a raster map representation
of the data. If the site file has category descriptions for all the sites
(and the -s flag is not used), the resulting map will reflect the categories
and their text descriptions (if any). This is why the sites list is read
twice - once to see if all sites have the categories and to put them in a
list, then once again to actually create the map. If the -s flag is used,
the first reading is by-passed.

Chris Rewerts, U.S. Army Construction Engineering Research Laboratory
rewerts@zorro.cecer.army.mil
May 20, 1993

-----------------------------------------------------------
UPDATE for average, sum, frequency, min and max value:

Jan 31, 2002
Marco Negretti
Politecnico di Milano - Campus Como
marco@geomatica.ing.unico.it
-----------------------------------------------------------

*/

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "gis.h"
#include "site.h"
#include "local_proto.h"
#include <stdio.h>

int legal_name (char *name)
{
   if (G_legal_filename(name) < 0)
   {
      fprintf (stderr, "\n%s: <%s> - illegal name\n",G_program_name (),name);
      G_usage ();
      exit(1);
   }
} /* END legal_name */



int main (int argc, char *argv[])
{
    FILE *fd;
    Site *s;
    struct Cell_head window;
    struct GModule *module;
    struct Categories cats;
    char *name, *mapset;
    char title_buf[1024];
    char *prev_cat;
    int quad_size;
    int quiet, retcode;
    int dims, strs, dbls;
    int said_it;
    int rows, cols, row;
    int num_index, str_index, rec_count, region_check;
    int scan_int, num_type;

    /* temp file */
    int fdtemp, fdtemp_fq, fdtemp_av, fdtemp_mx, fdtemp_mn;
    char *ntemp, *ntemp_fq, *ntemp_av, *ntemp_mx, *ntemp_mn;

    struct Option *input;
    struct Option *output;
    struct Option *size;
    struct Option *given_title;
    struct Option *field;
    struct Option *findex;
    struct Option *sindex;
    struct Flag   *verbose;
    struct Option *outfreq,*outavg,*outmax,*outmin;
    struct Flag   *zero_one;
    int freq,sum,media,min,max,check;

    void *rast, *rast_fq, *rast_av;
    char *str_fin, *cont_med;
    int med;
    int mediapunti;
    DCELL ritorno;

    RASTER_MAP_TYPE map_type, real_map_type;

    G_gisinit (argv[0]);

    module = G_define_module();
    module->description =        
                    "Converts a site file to a raster map";
                    
    input = G_define_option();
    input->key          ="input";
    input->description  ="Name of input site list";
    input->type		=TYPE_STRING;
    input->required     =YES;
    input->gisprompt 	="old,site_lists,site list";
 
    output = G_define_option();
    output->key         ="output";
    output->type        =TYPE_STRING;
    output->required    =NO;
    output->gisprompt   ="new,cell,raster";
    output->description ="Name of new raster file";

    outfreq = G_define_option();
    outfreq->key         ="frequencies";
    outfreq->type        =TYPE_STRING;
    outfreq->required    =NO;
    outfreq->gisprompt   ="new,cell,raster";
    outfreq->description ="Name of frequencies raster file";

    outavg = G_define_option();
    outavg->key         ="average";
    outavg->type        =TYPE_STRING;
    outavg->required    =NO;
    outavg->gisprompt   ="new,cell,raster";
    outavg->description ="Name of average raster file";

    outmax = G_define_option();
    outmax->key         ="maxima";
    outmax->type        =TYPE_STRING;
    outmax->required    =NO;
    outmax->gisprompt   ="new,cell,raster";
    outmax->description ="Name of maxima raster file";

    outmin = G_define_option();
    outmin->key         ="minima";
    outmin->type        =TYPE_STRING;
    outmin->required    =NO;
    outmin->gisprompt   ="new,cell,raster";
    outmin->description ="Name of minima raster file";

    size = G_define_option();
    size->key           ="size";
    size->description   ="Number of cells to surround site cell";
    size->type 		=TYPE_INTEGER;
    size->required      =NO;

    given_title = G_define_option ();
    given_title->key		="title";
    given_title->required	= NO;
    given_title->type		= TYPE_STRING;
    given_title->description	= "Title for the resulting raster map";
    
    field = G_define_option();
    field->key = "field";
    field->type = TYPE_STRING;
    field->required = NO;
    field->description = "Attribute field type to use for operation";
    field->options ="dim,decimal,cat";
    field->answer = "decimal";

    findex = G_define_option();
    findex->key = "findex";
    findex->type = TYPE_INTEGER;
    findex->required = NO;
    findex->description = "Attribute field number to use for operation";
    findex->answer = "1";
 
    sindex = G_define_option();
    sindex->key = "string";
    sindex->type = TYPE_INTEGER;
    sindex->required = NO;
    sindex->description = "String attribute number to use for description";
    sindex->answer = "1";

    zero_one = G_define_flag() ;
    zero_one->key         = 's' ;
    zero_one->description = "Create a single-valued (0/1) raster map";

    verbose = G_define_flag() ;
    verbose->key         = 'q' ;
    verbose->description = "Run quietly";

    if(G_parser(argc, argv))
        exit(1);
 
    /* inizializzazione variabili */
    quiet = verbose->answer;
    name = input->answer;
    sum=0;
    freq=0;
    media=0;
    max=0;
    min=0;
    check=0;

    if (!zero_one->answer)
    {
       if (output->answer)
       {
          legal_name(output->answer);
          check=1;
          sum=1;
       }
       if (outfreq->answer)
       {
          legal_name(outfreq->answer);
          check=1;
          freq=1;
       }
       if (outavg->answer)
       {
          legal_name(outavg->answer);
          check=1;
          media=1;
       }
       if (outmax->answer)
       {
          legal_name(outmax->answer);
          check=1;
          max=1;
       }
       if (outmin->answer)
       {
          legal_name(outmin->answer);
          check=1;
          min=1;
       }
    } /* END if(!zero_one->answer) */
    else if (output->answer)
    {
       legal_name(output->answer);
       check=1;
    }


    /* check output file name */
    if (!check)
    {
       fprintf(stderr,"\n%s: output file name???\n",G_program_name());
       exit(1);
    }


    if (given_title->answer)
        sprintf (title_buf, "%s", given_title->answer);

    /* INPUT DATA CHECK */
    if (size->answer)
    {
        if (sscanf (size->answer, "%d", &quad_size) != 1 || quad_size < 0)
        {
            fprintf(stderr,
            "\n%s: <%s> is an illegal size option.\n",
            G_program_name(), size->answer);
            G_usage();
            exit(1);
        }
    }
    else
        quad_size = 0;  /* our default - one cell per site */

    if (strncmp(field->answer,"dim",3) == 0 )
        num_type = SITE_COL_DIM;
    else if (strncmp(field->answer, "decimal", 7) == 0)
        num_type = SITE_COL_DBL;
    else if (strncmp(field->answer, "cat", 3) == 0)
        num_type = SITE_COL_NUL;
    else { /* Shouldn't happen */
       char msg[256];
       snprintf(msg, 256, "%s: \"%s\" is an unknown field type.\n",
                       G_program_name(), field->answer);
       G_fatal_error(msg);
    }

    /* controlla colonna dati */
    scan_int=sscanf(findex->answer,"%d",&num_index);
    if ((scan_int <= 0) || num_index < 1)
    {
        char msg[256];
        sprintf(msg,"%s: \"%s\" is an incorrect value for attribute field number.\n",
        G_program_name(), findex->answer );
        G_fatal_error (msg);
        exit(1);
    }
    num_index--;
    
    /* controlla colonna attributo */
    scan_int = sscanf(sindex->answer, "%d", &str_index);
    if ((scan_int <= 0) || str_index < 1)
    {
        char msg[256];
        sprintf(msg,"%s: \"%s\" is an incorrect value "
                        "for attribute field number.\n",
                        G_program_name(), findex->answer );
        G_fatal_error (msg);
        exit(1);
    }
    str_index--;
    G_get_window (&window);

    if (!quiet)
    {
        fprintf (stdout, "Using size option: %d\n", quad_size);
        if (zero_one->answer)
            fprintf (stdout, "Forcing single-valued (no-data/1) raster map\n");
        fprintf (stdout, "Finding and opening site list...\n");
    }

    mapset = G_find_sites(name, "");
    if (!mapset)
    {
	fprintf (stderr, "\n%s: site list <%s> not found.\n",
	G_program_name(), name);
	exit(1);
    }

    fd = G_fopen_sites_old (name, mapset);
    if (fd == NULL)
    {
        fprintf (stderr, "\n%s: could not open sites file <%s>",
	G_program_name(), name);
	exit(1);
    }
    if(G_site_describe (fd, &dims, &map_type, &strs, &dbls) != 0)
    {
	fprintf (stderr, "\n%s: wrong site format",G_program_name());
    }
    real_map_type = map_type;

    /* determina il tipo di raster */
    switch(num_type) {
        case SITE_COL_DBL:
            if(num_index >= dbls)
            {
                fprintf(stderr,"\n");
                G_fatal_error("selected decimal field column no. %d "
                        "not present in sites list.", num_index+1);
            }
            map_type = DCELL_TYPE;
            break;
        case SITE_COL_DIM:
            if(num_index >= dims || num_index < 2 )
            {
                fprintf(stderr,"\n");
                G_fatal_error("selected dimension no. %d "
                        "not present in sites list. or is a coordinate", num_index+1);
            }
            num_index -= 2;
            map_type = DCELL_TYPE;
            break;
        case SITE_COL_NUL:
            if(map_type < 0 && !(zero_one->answer))
            {
                fprintf(stderr, "\n");
                G_fatal_error("site_list does not have any category values");
            }
        default: /* Already caught.. */
        break;
    }
    
    if (str_index > strs) {
        G_warning("String attribute index out of range, resetting to default.");
        str_index = 0;
    }
    if(!quiet)
       fprintf (stderr, "\nSites map Type: %d, Dims: %d, Strs: %d, Dbls: %d\n", 
		       real_map_type, dims, strs, dbls);

    if (zero_one->answer)
    {
        map_type = CELL_TYPE;
    }

    /* alloca memoria per il site */
    s = G_site_new_struct (real_map_type, dims, strs, dbls); 

    /* apre i file temp */
    if((sum)||(media)||(zero_one->answer))
    {
       ntemp = G_tempfile();
       fdtemp = open (ntemp, O_CREAT | O_TRUNC | O_RDWR, 0660);
       if (fdtemp == -1)
          G_fatal_error ("Unable to create temporary file.");
    }
    else
       fdtemp = -1;

    if((freq)||(media))
    {
       ntemp_fq = G_tempfile();
       fdtemp_fq = open (ntemp_fq, O_CREAT | O_TRUNC | O_RDWR, 0660);
       if (fdtemp_fq == -1)
          G_fatal_error ("Unable to create temporary file.");
       rast_fq = G_allocate_raster_buf(CELL_TYPE);
       if (rast_fq == NULL)
          G_fatal_error ("Failed to allocate raster buffer.");
    }
    else
       fdtemp_fq = -1;

    if(media)
    {
       ntemp_av = G_tempfile();
       fdtemp_av = open (ntemp_av, O_CREAT | O_TRUNC | O_RDWR, 0660);
       if (fdtemp_av == -1)
          G_fatal_error ("Unable to create temporary file.");
       rast_av = G_allocate_raster_buf(DCELL_TYPE);
       if (rast_av == NULL)
          G_fatal_error ("Failed to allocate raster buffer.");
    }
    else
       fdtemp_av = -1;

    if(max)
    {
       ntemp_mx = G_tempfile();
       fdtemp_mx = open (ntemp_mx, O_CREAT | O_TRUNC | O_RDWR, 0660);
       if (fdtemp_mx == -1)
          G_fatal_error ("Unable to create temporary file.");
    }
    else
       fdtemp_mx = -1;

    if(min)
    {
       ntemp_mn = G_tempfile();
       fdtemp_mn = open (ntemp_mn, O_CREAT | O_TRUNC | O_RDWR, 0660);
       if (fdtemp_mn == -1)
          G_fatal_error ("Unable to create temporary file.");
    }
    else
       fdtemp_mn = -1;

    /* questo ci vuole sempre, indipendentemente dai falg iniziali */
    rast = G_allocate_raster_buf(map_type);
    if (rast == NULL)
        G_fatal_error ("Failed to allocate raster buffer.");

    rows = window.rows;
    cols = window.cols;

    if (rows < 1 || cols < 1)
        G_fatal_error ("Region setting has no cells");

    /*  zero out the entire file that will receive data   */
    if (!quiet)
        fprintf (stderr, "\ninput sites map: <%s> in <%s>\n", name, mapset);

    G_set_null_value(rast, cols, map_type);
    if (fdtemp_av!=-1)
       G_set_null_value(rast_av, cols, DCELL_TYPE);
    /* inizializza i file temporanei */
    for(row = 0; row < rows; row++)
    {
        if((sum)||(media)||(zero_one->answer))
        {
           if(write(fdtemp, rast, cols * G_raster_size(map_type))!=cols * G_raster_size(map_type))
              G_fatal_error("error while writing to temp file");
        }
        if((freq)||(media))
        {
           if(write(fdtemp_fq, rast_fq, cols * G_raster_size(CELL_TYPE))!=cols * G_raster_size(CELL_TYPE))
              G_fatal_error("error while writing to temp file");
        }
        if(media)
        {
           if(write(fdtemp_av, rast_av, cols * G_raster_size(DCELL_TYPE))!=cols * G_raster_size(DCELL_TYPE))
              G_fatal_error("error while writing to temp file");
        }
        if(max)
        {
           if(write(fdtemp_mx, rast, cols * G_raster_size(map_type))!=cols * G_raster_size(map_type))
              G_fatal_error("error while writing to temp file");
        }
        if(min)
        {
           if(write(fdtemp_mn, rast, cols * G_raster_size(map_type))!=cols * G_raster_size(map_type))
              G_fatal_error("error while writing to temp file");
        }
    }

    if (!quiet)
    {
        if ((sum)||(zero_one->answer))
           fprintf (stderr, "output raster map: <%s> in <%s>\n", output->answer, G_mapset());
        if (freq)
           fprintf (stderr, "output frequency raster map: <%s> in <%s>\n", outfreq->answer, G_mapset());
        if (media)
           fprintf (stderr, "output average raster map: <%s> in <%s>\n", outavg->answer, G_mapset());
        if (max)
           fprintf (stderr, "output average raster map: <%s> in <%s>\n", outmax->answer, G_mapset());
        if (min)
           fprintf (stderr, "output average raster map: <%s> in <%s>\n", outmin->answer, G_mapset());
    }

/* 
   if the site descriptions are all of the form: #n <label>
   then assign the site to the #n category with label <label>
   otherwise create a 0/1 cell file
*/
    if (!given_title->answer)
        sprintf (title_buf, "Created by %s from %s", G_program_name (), 
        G_fully_qualified_name (name, mapset) );
    G_init_raster_cats (title_buf, &cats);

    if (!quiet)
        fprintf (stdout, "transferring sites to raster file...\n");
    
    region_check = 0;
    rec_count = 1;
    said_it = 0;
    fseek (fd,0L,0);

    /* scansione del file site */
    while  ((retcode = G_site_get (fd, s))==0)
    {
        rec_count++;
        if(!G_site_in_region(s, &window))
        {
            region_check = 1;
        }
        if(!zero_one->answer)
        {
	    switch(num_type)
	    {
	        case SITE_COL_NUL:
		    switch(map_type) {
		        case CELL_TYPE:
		            G_set_raster_value_c(rast, s->ccat, map_type);
			    break;
			case FCELL_TYPE:
			    G_set_raster_value_f(rast, s->fcat, map_type);
			    break;
			case DCELL_TYPE:
			    G_set_raster_value_d(rast, s->dcat, map_type);
			    break;
		    }
		    break;
		case SITE_COL_DBL:
		    G_set_raster_value_d(rast, s->dbl_att[num_index], map_type);
		    break;
		case SITE_COL_DIM:
		    G_set_raster_value_d(rast, s->dim[num_index], map_type);
		    break;
		default: /* Should never happen */
		    break;
	    } /* end switch(num_type) */

            if(s->str_att)
	    {
                prev_cat = G_get_raster_cat(rast, &cats, map_type);
                if (strcmp(prev_cat, "") && strcmp(prev_cat,s->str_att[str_index]) && !quiet)
                {
                    if (!said_it)
                    {
                        fprintf (stderr,
                        "\nNOTE: Category description mismatch -\n");
                        fprintf (stderr,
                        "more than one label has been found for the same category:\n");
                        said_it = 1;
                    }
                    fprintf (stderr,
                    "    Cell Value:           %f\n", 
					G_get_raster_value_d(rast, map_type));
                    fprintf (stderr,
                    "    Previous label found: %s\n", prev_cat);
                    fprintf (stderr,
                    "    Current label found:  %s\n", s->str_att[str_index]);
                }
	        G_set_raster_cat (rast, rast, s->str_att[str_index], &cats, map_type);
		/*
		fprintf (stdout,"setting raster cats %f %s\n", *((FCELL *) rast), *s->str_att);
		*/
            } /* end if(s->str_att) */
        } /* end if(!zero_one->answer) */
        else /* zero_one->answer */
	    G_set_raster_value_c(rast, (CELL) 1, map_type);
        write_temp (fdtemp, fdtemp_fq, fdtemp_mx, fdtemp_mn, fdtemp_av, &window, s->east, s->north, quad_size, rast, map_type, zero_one->answer);
    }  /* END while((retcode = G_site_get (fd, s))==0) */

    if(!quiet)
    {
        switch (retcode)
        {
            case -1:
                fprintf(stderr, "Site list read successfully.\n");
                break;
            case 1:
                G_warning("Extra attributes found in a site record %d "
                                "Remaining data skipped", rec_count);
                break;
            case -2:
                G_warning("Error reading sites list at record %d. "
                        "Remaining data skipped.", rec_count);
                break;
            default: /* Unknown return code */
                G_warning("Got unknown return code: %d", retcode);
       } /* END switch(retcode) */
       if (region_check) /* Some sites out of region */
            G_warning("Some sites were outside the current region "
                    "and were not included.");
    }

    if(zero_one->answer)
    {
	G_set_cat((CELL ) 0, "no data", &cats);
	G_set_cat((CELL ) 1, "site data", &cats);
    }
	
    /* chiudo i file temp e scrivo raster */
    if ((sum)||(zero_one->answer))
    {
       close_temp(fdtemp, output->answer, ntemp, rows, cols, map_type);
       if (!quiet)
          fprintf (stdout, "creating support files ...\n");
       G_write_cats (output->answer, &cats);
       if (!quiet)
          fprintf(stdout, "\n<%s> raster file complete.\n", output->answer);
    }

    if (freq)
    {
       close_temp(fdtemp_fq, outfreq->answer, ntemp_fq, rows, cols, CELL_TYPE);
       if (!quiet)
          fprintf(stdout,"\n<%s> frequency raster file complete.\n", outfreq->answer);
    }
    if (media)
    {
       if (!sum)
          close(fdtemp);
       if (!freq)
          close(fdtemp_fq);
       close_temp(fdtemp_av, outavg->answer, ntemp_av, rows, cols, DCELL_TYPE);
       if (!sum)
          unlink(ntemp);
       if (!freq)
          unlink(ntemp_fq);
       if (!quiet)
       	  fprintf(stdout,"\n<%s> average raster file complete.\n", outavg->answer);
    }
    if (fdtemp_mx!=-1)
    {
       close_temp(fdtemp_mx, outmax->answer, ntemp_mx, rows, cols, map_type);
       if (!quiet)
          fprintf(stdout,"\n<%s> maxima raster file complete.\n", outmax->answer);
    }
    if (fdtemp_mn!=-1)
    {
       close_temp(fdtemp_mn, outmin->answer, ntemp_mn, rows, cols, map_type);
       if (!quiet)
       	fprintf(stdout,"\n<%s_m> minima raster file complete.\n", outmin->answer);
    }
}

